import platform
from enum import Enum
from functools import lru_cache
from os import environ

class Platform(Enum):
    LINUX = 'linux'
    MAC = 'darwin'
    WINDOWS = 'windows'
    ANDROID = 'android'

    @lru_cache()
    def __bool__(self):
        # android
        if self.value == self.ANDROID: 
            return 'ANDROID_STORAGE' in environ
        # mac, windows, linux
        return platform.system().lower() == self.value



import webbrowser
import threading
from mscan import settings
from mscan.scanner.interface import InterfaceManager, Interface
from mscan.scanner.scanner import Scanner
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.staticfiles import StaticFiles
from starlette.responses import FileResponse
import uvicorn
import socket
import asyncio

app = FastAPI()
scanner = Scanner()
imanager = InterfaceManager()

origins = ["*"]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

def local_ip():
    try:
        return socket.gethostbyname(socket.gethostname())
    except:
        pass



@app.get("/")
async def index_route():
    return FileResponse(settings.UI_BUILD_PATH / "index.html")


@app.get("/interfaces")
async def interfaces_route():
    """ Get list of interfaces """
    # exclude 255.0.0.0 netmask since it has too many ips, billions...
    interfaces = [
        i.to_dict() for i in imanager.get_interfaces()
        if i.prefix_length != 8
    ]
    return {"interfaces": interfaces, "default": imanager.get_default()}


@app.post("/scan")
async def scan_route(interface: Interface):
    res = await scanner.scan(interface)
    return res


app.mount("/", StaticFiles(directory=settings.UI_BUILD_PATH), name="public")


def open_in_browser(host):
    webbrowser.open(f'{settings.PROTOCOL}://{host}:{settings.PORT}')


def main():
    # interfaces = list(imanager.get_interfaces())
    # i = interfaces[3]
    # res = await scanner.scan(i)
    # print(res)
    # return
    host = settings.HOST
    if settings.HOST == '0.0.0.0':
        host = local_ip() or settings.HOST

    print(f"Mscan running at {settings.PROTOCOL}://{host}:{settings.PORT}/")
    threading.Timer(1.25, open_in_browser, (host, ) ).start()
    uvicorn.run(
        "main:app" if settings.DEV_MODE else app,
        host=settings.HOST,
        port=settings.PORT,
        reload=settings.DEV_MODE,
        workers=1,
        log_level=settings.LOG_LEVEL
    )


if __name__ == '__main__':
    main()
    




from icmplib import async_multiping


class MultiPinger:
    """ Multithreaded class for pinging hosts and display progress """

    @staticmethod
    async def ping(ips):
        return await async_multiping(ips, count=1, timeout=1, privileged=False)



from pathlib import Path
import argparse
import logging

# Command line arguments
parser = argparse.ArgumentParser("Mscan", description="Modern network scanner")
parser.add_argument("--port", default=8000, required=False, type=int)
parser.add_argument("--host", default="127.0.0.1", required=False)
args = parser.parse_args()

# General Settings
DEV_MODE = False
LOG_LEVEL = logging.INFO if DEV_MODE else logging.CRITICAL

# Server settings
PROTOCOL = 'http'
HOST = args.host
PORT = args.port

# UI settings
UI_BUILD_PATH = (Path(__file__).parent / ("../../ui/dist" if DEV_MODE else 'static')).resolve().absolute()
UI_BUILD_PATH.mkdir(exist_ok=True)



import socket
import psutil
from dataclasses import dataclass, asdict


@dataclass
class Interface:
    name: str
    address: str
    netmask: str
    family: int
    prefix_length: int

    @classmethod
    def from_psutil(cls, family, interface, snic):

        family = 6 if family == socket.AF_INET6 else 4
        netmask = snic.netmask
        if netmask:
            prefix_length = sum([bin(int(x)).count("1") for x in netmask.split(".")])
        else:
            prefix_length = None

        return cls(interface, snic.address, netmask, family, prefix_length)

    def to_dict(self):
        return {k: str(v) for k, v in asdict(self).items()}

    @classmethod
    def from_dict(cls, iface_dict):
        return cls(**iface_dict)


class InterfaceManager:

    @staticmethod
    def get_interfaces(family=socket.AF_INET):
        ifaces = psutil.net_if_addrs().items()
        for interface, snics in ifaces:
            for snic in snics:
                if snic.family == family:
                    yield Interface.from_psutil(family, interface, snic)

    def get_default(self, family=socket.AF_INET):
        ifaces = list(self.get_interfaces())

        s = socket.socket(family, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        default_if_ip = s.getsockname()[0]
        return next((i for i in ifaces if i.address == default_if_ip), ifaces[0])



import concurrent.futures
import socket
from dataclasses import dataclass, asdict
from typing import List
from mac_vendor_lookup import AsyncMacLookup
from .platform_detector import Platform
from .arp import Arp, Host as ArpHost
from .interface import Interface
from .multi_ping import MultiPinger
from .network import Network
import re

@dataclass
class LiveHost:
    ip: str
    mac: str
    vendor: str
    hostname: str

    def to_dict(self):
        return {k: str(v) for k, v in asdict(self).items()}


class Scanner:
    BLACKLIST_PREFIX = [str(i) for i in range(220, 265)]
    BLACKLIST_SUFFIX = ['255']

    @staticmethod
    def get_hostname(host):
        try:
            return socket.gethostbyaddr(host['ip'])[0]
        except:
            return ''

    @staticmethod
    def get_hostnames(hosts):
        with concurrent.futures.ThreadPoolExecutor(max_workers=2500) as executer:
            results = []
            for host, result in zip(hosts, executer.map(Scanner.get_hostname, hosts)):
                if result:
                    results.append({**host, 'hostname': result})
        return results

    @staticmethod
    def get_hostnames(hosts: List[ArpHost], max_workers=2500):

        def job(arp_host: ArpHost):
            try:
                hostname = socket.gethostbyaddr(arp_host.addr)[0]
            except:
                hostname = "unknown"
            return {"host": arp_host, "name": hostname}

        result = {}
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executer:
            futures = []
            for host in hosts:
                futures.append(executer.submit(job, host))
            for future in concurrent.futures.as_completed(futures):
                res = future.result()
                host, name = res.values()
                result[host.addr] = name
        return result

    @staticmethod
    async def get_vendor(mac: str):
        mac = mac.upper().replace('-', ':')
        try:
            vendor = await AsyncMacLookup().lookup(mac)
        except:
            return "unknown"
        return vendor

    @staticmethod
    async def get_live_hosts(hosts: List[ArpHost]):
        live_hosts = []
        hostnames = Scanner.get_hostnames(hosts)
        for host in hosts:
            suffix = host.addr[-3:]
            prefix = host.addr[:3]
            if prefix not in Scanner.BLACKLIST_PREFIX and suffix not in Scanner.BLACKLIST_SUFFIX:
                vendor = await Scanner.get_vendor(host.hwaddr)
                live_host = LiveHost(host.addr, host.hwaddr, vendor, hostnames.get(host.addr, "unknown"))
                live_hosts.append(live_host)
        return live_hosts


    def arp_scan(self, hosts):
        with concurrent.futures.ThreadPoolExecutor(max_workers=2500) as executer:
            results = []
            for host, result in zip(hosts, executer.map(Arp.get_mac, hosts)):
                if result:
                    result = re.sub(r'(..)', r'\1:', result)[:-1]
                    result = result.upper()
                    results.append(ArpHost(addr=host, hwaddr=result, addr_type='dynamic'))
        return results
    
    async def scan(self, iface: Interface) -> List[LiveHost]:
        network = Network.from_interface(iface)
        ips = [str(i) for i in network.ips]
        if Platform.WINDOWS:
            hosts = self.arp_scan(ips)
        else:
            await MultiPinger().ping(ips)
            hosts = Arp.get_tables(iface)
        live = await self.get_live_hosts(hosts)
        return live



import subprocess
from .platform_detector import Platform


class Ping:
    """Cross platform ping"""

    @staticmethod
    def ping(ip, count=1):
        count_flag = "-n" if Platform.WINDOWS else "-c"
        proc = subprocess.call(["ping", count_flag, str(count), ip], stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL)
        return proc == 0



import ipaddress
from dataclasses import dataclass

from .interface import Interface


@dataclass
class Network:
    default_gateway: ipaddress.IPv4Address
    ips: ipaddress.IPv4Network

    @classmethod
    def from_interface(cls, interface: Interface):
        network = ipaddress.IPv4Network(f'{interface.address}/{interface.prefix_length}', strict=False)
        default_gateway = network[1]
        return cls(default_gateway, network)



from .interface import Interface
from .platform_detector import Platform
import subprocess
from dataclasses import dataclass
from typing import List
import ctypes
from ipaddress import IPv4Address
@dataclass
class Host:
    addr: str
    hwaddr: str
    addr_type: str


class Arp:

    @staticmethod
    def parse(output, iface_ip) -> List[Host]:
        lines = output.strip().split("\n")
        lines = [l.strip() for l in lines if l.strip() != ""]
        tables = {}
        current = None
        for line in lines:
            if line.startswith('Interface'):
                current = line.split()[1]
                tables[current] = []
            elif "Physical" not in line:
                addr, hwaddr, addr_type = line.split()
                row = Host(addr, hwaddr, addr_type)
                tables[current].append(row)
        return tables.get(iface_ip)

    @staticmethod
    def get_tables(interface: Interface):
        if Platform.WINDOWS:
            proc = subprocess.run(["arp", "-a"], stdout=subprocess.PIPE)
            output = proc.stdout.decode("utf-8")
            tables = Arp.parse(output, interface.address)
            return tables or []
        elif Platform.LINUX:
            hosts = []
            with open('/proc/net/arp', 'r') as f:
                lines = f.read().splitlines()
                lines = lines[1:]
            for line in lines:
                ip, hw_type, flags, hw_addr, _mask, ifname = line.split()
                if ifname == interface.name and flags != "0x0":
                    hosts.append(Host(ip, hw_addr, hw_type))
            return hosts
        
    @staticmethod
    def get_mac(host):
        """ Send ARP who-has request using Microsoft's Iphlpapi.lib
        Args:
            host (str): IP address to request
        Returns:
            True if IP address answers to ARP who-has, False otherwise

        DWORD SendARP(
        _In_    IPAddr DestIP,
        _In_    IPAddr SrcIP,
        _Out_   PULONG pMacAddr,
        _Inout_ PULONG PhyAddrLen
        );

        DestIP is decimal value of inverted IP address, for instance :
        10.0.0.103 => 103.0.0.10 => 1728053258
        Microsoft recommands wsock32.inet_addr(), which I can't use with Python 3.
        """
        SendARP = ctypes.windll.Iphlpapi.SendARP
        #inetaddr = ctypes.windll.wsock32.inet_addr(host) #Fonctionne uniquement sous Python 2
        inetaddr = int(IPv4Address('.'.join(host.split(".")[::-1])))

        buffer = ctypes.c_buffer(6)
        addlen = ctypes.c_ulong(ctypes.sizeof(buffer))

        if SendARP(inetaddr, 0, ctypes.byref(buffer), ctypes.byref(addlen)) == 0:
            mac = buffer[:6]
            return mac.hex()
        return ''